/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    std::vector<Condition> conds_;              // 扫描条件
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<RecScan> scan_;

    SmManager *sm_manager_;

   public:
    IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string> index_col_names,
                    Context *context) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        conds_ = std::move(conds);
        // index_no_ = index_no;
        index_col_names_ = index_col_names; 
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        for (auto &cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                // lhs is on other table, now rhs must be on this table
                assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
                // swap lhs and rhs
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
        }
        fed_conds_ = conds_;
    }

    void beginTuple() override {
        checkRuntimeConds();

        //获取IxIndexHandle
        std::string index_name = sm_manager_->get_ix_manager()->get_index_name(tab_name_,index_col_names_);
        IxIndexHandle *ixIndexHandle = sm_manager_->ihs_.at(index_name).get();

        char low_key[index_meta_.col_tot_len];
        char up_key[index_meta_.col_tot_len];
        int offset = 0;
        int i = 0;
        for(;i<index_meta_.col_num;i++){
            if(!fillUpAndLowKeys(index_meta_.cols[i].name,low_key,up_key,offset,index_meta_.cols[i].len)){
                //是等值条件，这里可以直接跳出分析
                //以第一个等值条件作为查询条件
                break;
            }
            offset+=index_meta_.cols[i].len;
        }
        //遇到等值条件退出，用边界条件填充剩余的字段即可
        if(i!=index_meta_.col_num){
            offset+=index_meta_.cols[i].len;
            for(i++;i<index_meta_.col_num;i++){
                setKeyWithBoundValue(index_meta_.cols[i].name,low_key+offset,index_meta_.cols[i].len,true);
                setKeyWithBoundValue(index_meta_.cols[i].name,up_key+offset,index_meta_.cols[i].len,false);
                offset+=index_meta_.cols[i].len;
            }
        }
        index_meta_.printKey(low_key);
        index_meta_.printKey(up_key);

        //确认up_key大于low_key
        offset = 0;
        int compareResult = 0;
        for(size_t j = 0; j < index_meta_.cols.size(); j++){
            compareResult = ix_compare(low_key+offset,up_key+offset,index_meta_.cols[j].type,index_meta_.cols[j].len);
            if(compareResult!=0){
                break;
            }
            offset+=index_meta_.cols[j].len;
        }

        if(compareResult>0){
            //错误情况,low_key>up_key
            scan_ = std::make_unique<IxScan>(ixIndexHandle, ixIndexHandle->leaf_end(), ixIndexHandle->leaf_end(), sm_manager_->get_bpm());
        }else{
            scan_ = std::make_unique<IxScan>(ixIndexHandle, ixIndexHandle->lower_bound(low_key), ixIndexHandle->upper_bound(up_key), sm_manager_->get_bpm());
        }

        while (!scan_->is_end()) {
            rid_ = scan_->rid();
            auto record = fh_->get_record(rid_, context_);
            if (isSatifiedConds(cols_, fed_conds_, *record)){
                //找到满足条件的记录
                break;
            }
            scan_->next();
        }
    }

    void nextTuple() override {
        checkRuntimeConds();
        scan_->next();
        while (!scan_->is_end()) {
            rid_ = scan_->rid();
            auto record = fh_->get_record(rid_, context_);
            if (isSatifiedConds(cols_, fed_conds_, *record)){
                //找到满足条件的记录
                break;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if (scan_->is_end())
            return nullptr;
        else
            return fh_->get_record(rid_, context_);
    }

    Rid &rid() override { return rid_; }

    /**
     * @description: 用于填充low_key和up_key
     * @param col_name 列名
     * @param low_key
     * @param up_key
     * @param offset 在index key中的偏移
     * @param len 字段长度
     * @return 是否是等值条件，等值条件直接返回true，否则返回false
     * 注：对于单个字段，只能有一个等于条件，这里应该在语义分析解决？
     */
    bool fillUpAndLowKeys(std::string col_name,char *low_key,char *up_key, int offset, int len){
        bool set_low_key = false;
        bool set_up_key = false;
        char temp_low_key[len];
        char temp_up_key[len];
        //对于每一个索引字段，遍历所有条件
        for(auto &cond: conds_){
            //列名相同，且条件右值是常量
            if(cond.lhs_col.col_name == col_name && cond.is_rhs_val){
                if(cond.op == OP_EQ){
                    if(set_low_key&&
                       ix_compare(temp_low_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)>0){
                        //错误情况
                    }
                    if(set_up_key&&
                       ix_compare(temp_up_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)<0){
                        //错误情况
                    }
                    memcpy(up_key + offset,cond.rhs_val.raw->data,len);
                    memcpy(low_key + offset,cond.rhs_val.raw->data,len);
                    // 对于一个字段只认为有一个等于条件，所以直接return
                    return true;
                }else if(cond.op == OP_GE){
                    if(set_up_key&&
                        ix_compare(temp_up_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)<0) {
                        //错误情况
                    }
                    if(set_low_key&&
                        ix_compare(temp_low_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)<=0){
                        //无需更新的情况
                        continue;
                    }
                    memcpy(temp_low_key,cond.rhs_val.raw->data,len);
                    if(!set_up_key){
                        setKeyWithBoundValue(col_name,temp_up_key,len,false);
                    }
                    set_low_key=true;
                }else if(cond.op == OP_GT){
                    if(set_up_key&&
                       ix_compare(temp_up_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)<=0) {
                        //错误情况
                    }
                    if(set_low_key&&
                       ix_compare(temp_low_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)<=0){
                        //无需更新的情况
                        continue;
                    }
                    memcpy(temp_low_key,cond.rhs_val.raw->data,len);
                    if(!set_up_key){
                        setKeyWithBoundValue(col_name,temp_up_key,len,false);
                    }
                    set_low_key=true;
                }
                else if(cond.op == OP_LE){
                    if(set_low_key&&
                       ix_compare(temp_low_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)>0) {
                        //错误情况
                    }
                    if(set_up_key&&
                       ix_compare(temp_up_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)>=0){
                        //无需更新的情况
                        continue;
                    }
                    memcpy(temp_up_key,cond.rhs_val.raw->data,len);
                    if(!set_low_key){
                        setKeyWithBoundValue(col_name,temp_low_key,len,true);
                    }
                    set_up_key=true;
                }
                else if(cond.op == OP_LT){
                    if(set_low_key&&
                       ix_compare(temp_low_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)>=0) {
                        //错误情况
                    }
                    if(set_up_key&&
                       ix_compare(temp_up_key,cond.rhs_val.raw->data,cond.rhs_val.type,len)>=0){
                        //无需更新的情况
                        continue;
                    }
                    memcpy(temp_up_key,cond.rhs_val.raw->data,len);
                    if(!set_low_key){
                        setKeyWithBoundValue(col_name,temp_low_key,len,true);
                    }
                    set_up_key=true;
                }
            }
        }

        memcpy(low_key+offset,temp_low_key,len);
        memcpy(up_key+offset,temp_up_key,len);
        return false;
    }

    /**
     * 将low_key或者up_key用边界值填充
     * @param colName 字段名
     * @param key
     * @param len
     * @param isLowKey 是否是low_key
     */
    void setKeyWithBoundValue(std::string colName,char *key,int len,bool isLowKey){
        switch(tab_.get_col(colName)->type){
            case TYPE_INT:
                if(isLowKey){
                    memcpy(key,&INT_MIN_,len);
                }else{
                    memcpy(key,&INT_MAX_,len);
                }
                break;
            case TYPE_FLOAT:
                if(isLowKey){
                    memcpy(key,&FLOAT_MIN,len);
                }else{
                    memcpy(key,&FLOAT_MAX,len);
                }
                break;
            case TYPE_STRING:
                if(isLowKey){
                    memset(key,0,len);
                }else{
                    memset(key,255,len);
                }
                break;
        }

    }


    /**
     * 检查运行时条件
     */
    void checkRuntimeConds() {
        for (auto &cond : fed_conds_) {
            assert(cond.lhs_col.tab_name == tab_name_);
            if (!cond.is_rhs_val) {
                assert(cond.rhs_col.tab_name == tab_name_);
            }
        }
    }

    bool isSatisfiedCond(const std::vector<ColMeta> &cols, const Condition &condition, const RmRecord *record) {
        auto lhs_col = get_col(cols, condition.lhs_col);
        //左值，对应的列/字段值
        char *lhs_key = record->data + lhs_col->offset;
        //右值，条件
        char *rhs_key;
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;
        if(condition.is_rhs_val) {
            rhs_key = condition.rhs_val.raw->data;
            rhs_type = condition.rhs_val.type;
        }else{
            auto rhs_col = get_col(cols, condition.rhs_col);
            rhs_key = record->data + rhs_col->offset;
            rhs_type = rhs_col->type;
        }
        if(lhs_type != rhs_type) {
            //类型不匹配
            assert(false);
            return false;
        }
        int compareResult = ix_compare(lhs_key, rhs_key, lhs_type, lhs_col->len);
        switch(condition.op){
            case OP_EQ:
                return compareResult == 0;
            case OP_NE:
                return compareResult != 0;
            case OP_LT:
                return compareResult < 0;
            case OP_GT:
                return compareResult > 0;
            case OP_LE:
                return compareResult <= 0;
            case OP_GE:
                return compareResult >= 0;
            default:
                throw InternalError("Unexpected op type");
        }
    }

    bool isSatifiedConds(const std::vector<ColMeta> &cols,const std::vector<Condition> &conds,const RmRecord record){
        for(auto cond:conds){
            if(!isSatisfiedCond(cols,cond,&record)){
                return false;
            }
        }
        return true;
    }
    size_t tupleLen() const override { return len_; }

    bool is_end() const override { return scan_->is_end(); }

    const std::vector<ColMeta> &cols() const override { return cols_; }

    ColMeta get_col_offset(const TabCol &target) {
        for (auto col : cols_) {
            if (target.col_name == col.name) {
                return col;
            }
        }
        return ColMeta();
    };

};