/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }
    std::unique_ptr<RmRecord> Next() override {
        // 检查类型
        for (auto &set_clause : set_clauses_) {
            auto lhs_col = tab_.get_col(set_clause.lhs.col_name);
            //INT类型可转换为FLOAT
            if (lhs_col->type == TYPE_FLOAT && set_clause.rhs.type == TYPE_INT) {
                set_clause.rhs.type = TYPE_FLOAT;
                set_clause.rhs.float_val = set_clause.rhs.int_val;
            } 
            else if (lhs_col->type != set_clause.rhs.type) {
                throw IncompatibleTypeError(coltype2str(lhs_col->type), coltype2str(set_clause.rhs.type));
            }

            set_clause.rhs.init_raw(lhs_col->len);                
        }
        
        for(auto &rid : rids_){
            // 获取源数据
            std::unique_ptr<RmRecord> rec = fh_->get_record(rid, context_);
            RmRecord old_value = *rec;

            // 构造新数据
            for (size_t i = 0; i < set_clauses_.size(); i++) {
                auto col = tab_.get_col(set_clauses_[i].lhs.col_name);
                memcpy(rec->data + col->offset, set_clauses_[i].rhs.raw->data, col->len);
            }             
            RmRecord new_value = *rec;

            // 同步更新索引
            // 首先需要唯一性检查
            // 遍历所有存在的索引，如果索引中已经存在新数据，抛出异常
            for(auto& index:tab_.indexes){
                char newKey[index.col_tot_len];
                char oldKey[index.col_tot_len];
                index.getKey(&old_value,oldKey);
                index.getKey(&new_value,newKey);
                //当索引值不同时才进行比较
                if(memcmp(newKey,oldKey,index.col_tot_len)!=0){
                    auto &indexHanle = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols));
                    if(indexHanle->find(newKey,context_->txn_)){
                        throw UniqueConflictError(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols),new_value.data);
                    }
                }

            }

            // 更新索引
            // 遍历所有存在的索引，更新索引
            for(auto& index:tab_.indexes){
                char newKey[index.col_tot_len];
                char oldKey[index.col_tot_len];
                index.getKey(&old_value,oldKey);
                index.getKey(&new_value,newKey);
                auto &indexHanle = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols));
                if(memcmp(newKey,oldKey,index.col_tot_len)!=0){

                    indexHanle->delete_entry(oldKey,context_->txn_);
                    indexHanle->insert_entry(newKey,rid,context_->txn_);
                }
                indexHanle->print_tree();
            }

            // 更新数据
            fh_->update_record(rid, rec->data, context_);

        }
        return nullptr;
    }

    Rid &rid() override { return _abstract_rid; }
};