/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "sm_manager.h"

#include <sys/stat.h>
#include <unistd.h>

#include <fstream>

#include "index/ix.h"
#include "record/rm.h"
#include "record_printer.h"

/**
 * @description: 判断是否为一个文件夹
 * @return {bool} 返回是否为一个文件夹
 * @param {string&} db_name 数据库文件名称，与文件夹同名
 */
bool SmManager::is_dir(const std::string& db_name) {
    struct stat st;
    return stat(db_name.c_str(), &st) == 0 && S_ISDIR(st.st_mode);
}

/**
 * @description: 创建数据库，所有的数据库相关文件都放在数据库同名文件夹下
 * @param {string&} db_name 数据库名称
 */
void SmManager::create_db(const std::string& db_name) {
    if (is_dir(db_name)) {
        throw DatabaseExistsError(db_name);
    }
    std::cout<<"DB\n\n\n";
    //为数据库创建一个子目录
    std::string cmd = "mkdir " + db_name;
    if (system(cmd.c_str()) < 0) {  // 创建一个名为db_name的目录
        throw UnixError();
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    //创建系统目录
    DbMeta *new_db = new DbMeta();
    new_db->name_ = db_name;

    // 注意，此处ofstream会在当前目录创建(如果没有此文件先创建)和打开一个名为DB_META_NAME的文件
    std::ofstream ofs(DB_META_NAME);

    // 将new_db中的信息，按照定义好的operator<<操作符，写入到ofs打开的DB_META_NAME文件中
    ofs << *new_db;  // 注意：此处重载了操作符<<

    delete new_db;

    // 创建日志文件
    disk_manager_->create_file(LOG_FILE_NAME);

    // 回到根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 删除数据库，同时需要清空相关文件以及数据库同名文件夹
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::drop_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    std::string cmd = "rm -r " + db_name;
    if (system(cmd.c_str()) < 0) {
        throw UnixError();
    }
    flush_meta();
}

/**
 * @description: 打开数据库，找到数据库对应的文件夹，并加载数据库元数据和相关文件
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::open_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }

    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }

    //打开一个名为DB_META_NAME的文件
    std::ifstream ifs(DB_META_NAME);

    //读取数据库元数据
    ifs>>db_;
    ifs.close();
    //打开所有的表文件
    for (auto &entry : db_.tabs_) {
        auto &tab = entry.second;
        fhs_.emplace(tab.name, rm_manager_->open_file(tab.name));
        // 打开表对应的索引文件
        for(auto &index:tab.indexes) {
            ihs_.emplace(ix_manager_->get_index_name(tab.name, index.cols),
                         ix_manager_->open_index(tab.name, index.cols));
        }
    }
}

/**
 * @description: 把数据库相关的元数据刷入磁盘中
 */
void SmManager::flush_meta() {
    // 默认清空文件
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
}

/**
 * @description: 关闭数据库并把数据落盘
 */
void SmManager::close_db() {
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
    db_.name_.clear();
    db_.tabs_.clear();

    //写回记录
    for( auto &entry:fhs_ ){
        rm_manager_->close_file( entry.second.get() );

    }
    //写回所有索引
    for( auto &entry:ihs_ ){
        ix_manager_->close_index( entry.second.get() );
    }



    fhs_.clear();

    //退出数据库目录
    if (chdir("..") < 0)
        throw UnixError();
    flush_meta();
}   

/**
 * @description: 显示所有的表,通过测试需要将其结果写入到output.txt,详情看题目文档
 * @param {Context*} context 
 */
void SmManager::show_tables(Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    outfile << "| Tables |\n";
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"Tables"}, context);
    printer.print_separator(context);
    for (auto &entry : db_.tabs_) {
        auto &tab = entry.second;
        printer.print_record({tab.name}, context);
        outfile << "| " << tab.name << " |\n";
    }
    printer.print_separator(context);
    outfile.close();
}

/**
 * @description: 显示表的元数据
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::desc_table(const std::string& tab_name, Context* context) {
    TabMeta &tab = db_.get_table(tab_name);

    std::vector<std::string> captions = {"Field", "Type", "Index"};
    RecordPrinter printer(captions.size());
    // Print header
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);
    // Print fields
    for (auto &col : tab.cols) {
        std::vector<std::string> field_info = {col.name, coltype2str(col.type), col.index ? "YES" : "NO"};
        printer.print_record(field_info, context);
    }
    // Print footer
    printer.print_separator(context);
}

/**
 * @description: 创建表
 * @param {string&} tab_name 表的名称
 * @param {vector<ColDef>&} col_defs 表的字段
 * @param {Context*} context 
 */
void SmManager::create_table(const std::string& tab_name, const std::vector<ColDef>& col_defs, Context* context) {
    if (db_.is_table(tab_name)) {
        throw TableExistsError(tab_name);
    }
    // Create table meta
    int curr_offset = 0;
    TabMeta tab;
    tab.name = tab_name;
    for (auto &col_def : col_defs) {
        ColMeta col = {.tab_name = tab_name,
                       .name = col_def.name,
                       .type = col_def.type,
                       .len = col_def.len,
                       .offset = curr_offset,
                       .index = false};
        curr_offset += col_def.len;
        tab.cols.push_back(col);
    }
    // Create & open record file
    int record_size = curr_offset;  // record_size就是col meta所占的大小（表的元数据也是以记录的形式进行存储的）
    rm_manager_->create_file(tab_name, record_size);
    db_.tabs_[tab_name] = tab;
    // fhs_[tab_name] = rm_manager_->open_file(tab_name);
    fhs_.emplace(tab_name, rm_manager_->open_file(tab_name));

    flush_meta();
}

/**
 * @description: 删除表7
 * @param {string&} tab_name 表的名称
 * @param {Context*} context
 */
void SmManager::drop_table(const std::string& tab_name, Context* context) {
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    //删除表的数据文件
    rm_manager_->close_file( fhs_.at(tab_name).get());
    rm_manager_->destroy_file(tab_name);

    //删除表的所有索引
    for(auto index:db_.get_table(tab_name).indexes){
        drop_index(tab_name,index.cols,context);
    }
    db_.get_table(tab_name).indexes.clear();

    //从数据库元数据中删除表项
    db_.tabs_.erase(tab_name);
    //从库中哈希表中删除该表项
    fhs_.erase(tab_name);

    //更新元数据
    flush_meta();
}

/**
 * @description: 创建索引
 * @param {string&} tab_name 表的名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::create_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    //判断索引是否已经存在
    if(ix_manager_->exists(tab_name, col_names)) {
        throw IndexExistsError(tab_name, col_names);
    }
    // 索引不存在

    // 获取表元数据
    TabMeta & tabMeta = db_.get_table(tab_name);
    // 创建索引，调用ix_manager的create_index函数
    int col_tot_len = 0;
    // 构造列元数据，同时计算出索引长度
    std::vector<ColMeta> cols;
    for(auto &col_name : col_names) {
        auto col = std::find_if(tabMeta.cols.begin(), tabMeta.cols.end(), [&](const ColMeta &col) { return col.name == col_name; });
        if(col == tabMeta.cols.end()) {
            throw ColumnNotFoundError(col_name);
        }
        cols.push_back(*col);
        col_tot_len += col->len;
    }


    //打开表文件
    RmFileHandle fileHandle = *fhs_.at(tab_name).get();
    //一致性检查
    std::vector<char *> keys;
    for(RmScan scan(&fileHandle);!scan.is_end();scan.next()){
        char buf[col_tot_len];
        auto record = fileHandle.get_record(scan.rid(),context);
        int offset = 0;
        for(auto &col : cols){
            std::memcpy(buf+offset,record->data+col.offset,col.len);
            offset += col.len;
        }
        for(auto key:keys){
            if(std::memcmp(key,buf,col_tot_len)==0){
                throw UniqueConflictError(ix_manager_->get_index_name(tab_name, col_names),buf);
            }
            keys.push_back(buf);
        }
    }
    // 创建索引文件
    ix_manager_->create_index(tab_name, cols);

    // 插入所有的记录到索引中
    //打开索引文件
    std::unique_ptr<IxIndexHandle>  ixIndexHandle = ix_manager_->open_index(tab_name,cols);
    //单条索引的缓冲区
    ixIndexHandle->print_tree();
    char buf[col_tot_len];
    //遍历表记录插入记录
    for(RmScan scan(&fileHandle);!scan.is_end();scan.next()){
        auto record = fileHandle.get_record(scan.rid(),context);
        int offset = 0;
        for(auto &col : cols){
            std::memcpy(buf+offset,record->data+col.offset,col.len);
            offset += col.len;
        }
        ixIndexHandle->insert_entry(buf,scan.rid(),context->txn_);
        ixIndexHandle->print_tree();
    }
    //更新表元数据
    tabMeta.indexes.push_back({tab_name, col_tot_len, (int)cols.size(), cols});

    //更新系统管理器的索引元数据
    ihs_.emplace(ix_manager_->get_index_name(tab_name, col_names), std::move(ixIndexHandle));

    //将数据库元数据刷入磁盘
    flush_meta();

}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    //判断索引是否存在
    if(!ix_manager_->exists(tab_name, col_names)) {
        throw IndexNotFoundError(tab_name, col_names);
    }
    // 索引存在
    // 获取表元数据
    TabMeta & tabMeta = db_.get_table(tab_name);
    //删除索引，调用ix_manager的drop_index函数
    std::string indexName = ix_manager_->get_index_name(tab_name, col_names);
    // 先关闭索引
    ix_manager_->close_index(ihs_.at(indexName).get());
    // 删除索引文件
    ix_manager_->destroy_index(tab_name, col_names);

    // 更新表元数据
    tabMeta.indexes.erase(tabMeta.get_index_meta(col_names));
    //更新系统管理器的索引元数据
    ihs_.erase(indexName);

    flush_meta();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<ColMeta>&} 索引包含的字段元数据
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<ColMeta>& cols, Context* context) {
    std::vector<std::string> col_names;
    for(auto &col:cols) {
        col_names.push_back(col.name);
    }
    drop_index(tab_name, col_names, context);
}

/**
 * @description: 展示索引
 * @param {string&} tab_name 表名称
 * @param {Context*} context
 */
void SmManager::show_index(std::string tab_name, Context *context) {
    //获取表元数据
    TabMeta& tabMeta = db_.get_table(tab_name);

    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"index"}, context);
    printer.print_separator(context);
    for (auto &index : tabMeta.indexes) {
        outfile<<"| "<<tab_name<<" | unique | (";
        outfile<<index.cols[0].name;
        for(int i=1;i<index.cols.size();i++){
            outfile<<","<<index.cols[i].name;
        }
        outfile<<") |\n";
        printer.print_record({ix_manager_->get_index_name(tab_name, index.cols)},context);
    }
    printer.print_separator(context);
    outfile.close();
}
